"""
PyGL3Display
(C) David Griffin (2010-present)
"Habilain" is a pseudonym used by David Griffin in a number of places.

A high performance Pygame OpenGL library, capable of handling 1000s of sprites

Requirements: Python, Pygame, PyOpenGL, Open GL 2.0 graphics hardware

Licensed under GPLv2 or GPLv3. You should have received a copy of both of these
licenses in the files GPLv2.txt and GPLv3.txt. If not these licenses can be
found at http://www.gnu.org/licenses/gpl-2.0.html (GPLv2) and
http://www.gnu.org/licenses/gpl-3.0.html (GPLv3). By using this file you agree
to be bound by at least one of these licenses.

Other licenses, including commercial/proprietary licenses, can be arranged by 
contacting me via e-mail at habilain@gmail.com. If you do not have in physical 
writing an alternative license agreement from me, you must use one of the 
licenses specified above when releasing work which makes use of this library.

pygl3display.compute.dumbtriangle

A factory for creating 'dumb triangles' which are used extensively in GPU
compute (where the shader is the bit that does the interesting stuff).
"""

from pygl3display.twodbases import TwoDThing
from pygl3display.vertexarrays import VertexArraySpec
from pygl3display.twodshaders import TwoDShader, TwoDShaderCommon
from pygl3display.glcommoninst import GLCommon
from pygl3display.rendertargets import TextureTarget

class Process(object):
    def __init__(self, makeRenderer, triangleClass):
        #self.renderer = renderer
        self.makeRenderer = makeRenderer
        self.triangleClass = triangleClass
        
    def __getattr__(self, key):
        if key == 'renderer':
            self.renderer = self.makeRenderer()
            return self.renderer
        else:
            return super(Process, self).__getattribute__(key)
        
    def render(self, targets, size):
        target = TextureTarget(targets, size, compute = True)
        triangles = self.triangleClass(((0, 0), (0, size), (size, 0), 
                        (size, 0), (size, size), (0, size)), target=target)
        GLCommon.doTasks(target)
        return target.textures

        
        
def makeDumbTriangle(name, commoninst):
    class DumbTriangles(TwoDThing):
        """The class that represents a group of triangles drawn to screen"""
        
        def __init__(self, vertices, pos = (0, 0),
                     layer = 0, target = None, shader = None):
            """Initialise the NoisyTriangles object:
            Args:
              vertices: A list of coordinates relative to the origin. 
                        NoisyTriangles will be drawn between these coordinates.
                        Note: Whilst the positions of each vertex can be changed,
                        the number of points cannot be changed
              pos: The position on screen to draw the origin point to
              layer: The layer the triangle should render on (default: 0)
              target: A render target for using FBOs; defaults to None
             Advanced:
              shader: The shader the triangle should use. Default to shared manager
            """
            if shader is not None:
                self._shader = shader
            else:
                self._shader = commoninst.defaultShader
            
            verticesLen = len(vertices)    
            if verticesLen % 3:
                raise ValueError("Need to pass in a multiple of 3 vertices" +
                                 "for triangles")

            self._layer = layer
            self._target = target
            
            renderLen = verticesLen
                
            self._renderer = self._shader.getRenderer()
            self._vertexArrays = self._renderer.getVertexArrays(verticesLen)
            self._slots = self._vertexArrays.getDataSlots(verticesLen)
            self._rslots = self._renderer.getRenderSlots(renderLen, 
                                self._layer, self._target, self._vertexArrays)
            
            self._rdata = self._slots
            self._rslots.upload(self._rdata)
            self.__verticesLen = verticesLen
            self.__renderLen = renderLen
            
            self.setVertices(vertices)
            self.setPosition(pos)
                        
        setVertices = TwoDThing.setVertices
        setLayer = TwoDThing.setLayer
        setVertex = TwoDThing.setVertex
    DumbTriangles.__name__ = name + 'Triangles'
    return DumbTriangles
    
def makeDumbTriangleCommon(name, shaderClass):
    class DumbTriangleCommonClass(object):
        """The shared state between 2d triangles"""
        def __init__(self):
            self.spec = VertexArraySpec(name+'Arrays')
            self.spec.addArray('pos')

            self.spec.addAttrib('pos', 'Position', 2, 0)
            self.spec.addAttrib('pos', 'Offset', 2, 2, True)

        def __getattr__(self, attr):
            if attr == 'defaultShader':
                self.makeDefaultShader()
                return self.defaultShader
            else:
                raise AttributeError('NoisyTriangleCommon does not have ' + attr)
                
        def makeDefaultShader(self):
            """Create the default shader object"""
            self.defaultShader = shaderClass(self)
    DumbTriangleCommonClass.__name__ = name + 'CommonClass'
    return DumbTriangleCommonClass()
    
def makeDumbTriangleShader(name, fragprog, renderer):
    class DumbTriangleShader(TwoDShader):
        """A shader for 2d triangles"""
        
        def __init__(self, common):
            """Initialise the 2d Triangle Shader.
            Args:
             arrays: a VertexArrays object that contains render data
             cameraObj: A TwoDCameraManager object for camera state. Default to shared
                        camera state."""
                        
            vertexProg = """
                vec2 applyTransform(in vec2 pos, in vec2 origin, in vec2 offset, 
                            in float xScale, in float yScale, in float rotation);
                attribute vec4 pos;   
                uniform mat4 projMatrix;
                varying vec2 texpos;
                void main()
                {   
                    vec2 transform = applyTransform(pos.xy, vec2(0.0, 0.0), 
                                     pos.zw, 1.0, 1.0, 0.0);
                    gl_Position = projMatrix * vec4(transform.xy, 0.0, 1.0);
                    texpos = (gl_Position.xy + 1.0) / 2.0;
                }"""
            
            fragmentProgs = []
            if isinstance(fragprog, list):
                fragmentProgs.extend(fragprog)
            else:
                fragmentProgs.append(fragprog)
            
            super(DumbTriangleShader, self).__init__(
                vertexProgs= [TwoDShaderCommon.applyTransformSource, 
                vertexProg], fragmentProgs=fragmentProgs)
                
            self.triangleRenderer = renderer(self, common)

        def getRenderer(self):
            """Get the appropriate renderer for triangles"""
            return self.triangleRenderer
            
    DumbTriangleShader.__name__ = name + 'Shader'
    return DumbTriangleShader
    
#def makeDumbTriangleAPI(name, commonClass):
#    class DumbTriangleAPI(object):
#        def __init__(self):
#            self.__renderer = commonClass.defaultShader.getRenderer()
#            
#        def __getattr__(self, key):
#            return getattr(self.__renderer, key)
#    DumbTriangleAPI.__name__ = name + 'API'
#    return DumbTriangleAPI 
    
def makeDumbTriangles(name, fragmentProg, renderClass, processClass, 
                      call=None):
    ShaderClass = makeDumbTriangleShader(name, fragmentProg, renderClass)
    common = makeDumbTriangleCommon(name, ShaderClass)
    TriangleClass = makeDumbTriangle(name, common)
    process = processClass(lambda: common.defaultShader.getRenderer(), 
                           TriangleClass)
    if call is None:
        # See if it's possible to guess the call method.
        # If there is only 1 non-init method specified on the process,
        # then it's that. Otherwise, it needs to be given by user.
        candidate = None
        for name in type(process).__dict__:
            thing = getattr(process, name)
            if not name.startswith('_') and callable(thing):
                if not candidate:
                    candidate = thing
                else:
                    candidate = None
                    break
    else:
        candidate = getattr(process, name)
    
    class APIClass(object):
        def __init__(self):
            self.__process = process
        
        def __getattr__(self, key):
            return getattr(self.__process, key)
        
        if candidate:
            __call__ = candidate
            
    APIClass.__name__ = name + 'API'
    
    return APIClass()

