"""
PyGL3Display
(C) David Griffin (2010-present)
"Habilain" is a pseudonym used by David Griffin in a number of places.

A high performance Pygame OpenGL library, capable of handling 1000s of sprites

Requirements: Python, Pygame, PyOpenGL, Open GL 2.0 graphics hardware

Licensed under GPLv2 or GPLv3. You should have received a copy of both of these
licenses in the files GPLv2.txt and GPLv3.txt. If not these licenses can be
found at http://www.gnu.org/licenses/gpl-2.0.html (GPLv2) and
http://www.gnu.org/licenses/gpl-3.0.html (GPLv3). By using this file you agree
to be bound by at least one of these licenses.

Other licenses, including commercial/proprietary licenses, can be arranged by 
contacting me via e-mail at habilain@gmail.com. If you do not have in physical 
writing an alternative license agreement from me, you must use one of the 
licenses specified above when releasing work which makes use of this library.

pygl3display.examples.platformer

This file implements a platforming demo, which may become a full game in the
future.
"""
from __future__ import division

#Import Modules
# Magic so that examples work if trying from an extracted archive
try:
    import pygl3display
except ImportError:
    import sys, os.path
    sys.path.append(os.path.join('..', '..'))
    import pygl3display
pygl3display.configure('pygame')
from pygl3display import *
import os, os.path, random, sys
import pygame
from pygame.locals import (KEYUP, KEYDOWN, K_LEFT, K_UP, K_RIGHT, QUIT, 
                           K_ESCAPE)
from collections import namedtuple
from math import copysign

DEBUG = False

class BroadcastSet(set):
    """A set of functions which should all have the same function signature.
    When broadcast is called, all functions inside are called"""
    def broadcast(self, *args, **kwargs):
        """Broadcast to all functions in the set"""
        for callback in self: 
            callback(*args, **kwargs)
        
class Broadcaster(dict):
    """A broadcasting system in native types. Usage:
    broadcaster[signal].add(function): add function to listening to signal
    broadcaster[signal].broadcast(args): broadcast to everything listening
                                         to signal"""
    def __missing__(self, key):
        """Automatically create the broadcast sets for new signals"""
        self[key] = BroadcastSet()
        return self[key]

broadcastSystem = Broadcaster()

class ScheduledTask(object):
    """A deferred call to a function"""
    def __init__(self, function, args=[], kwargs={}):
        """Initialise the deferred task.
        Args:
         function: function to call
         args: a list of arguments to the function
         kwargs: a dict of kwargs to the function"""
        self.function = function
        self.args = args
        self.kwargs = kwargs
        
    def __call__(self):
        """Syntactic sugar"""
        self.function(*self.args, **self.kwargs)

class Scheduler(dict):
    def __init__(self, current, mod, broadcastSystem=broadcastSystem, 
                 *args, **kwargs):
        super(Scheduler, self).__init__(*args, **kwargs)
        self.current = current
        self.mod = mod
        broadcastSystem['tick'].add(self.tick)
        
    """We need hashing to be able to receive a broadcast, but 
    dicts in general are not hashable. As we don't care about the nice
    properties of a hash in this case, we can just take the hash function of
    object."""
    __hash__ = object.__hash__
        
    def __getitem__(self, key):
        if isinstance(key, (int, float)):
            key = ((key + self.current) % self.mod,)
        return super(Scheduler, self).__getitem__(key)
            
    def __missing__(self, key):
        if isinstance(key, tuple):
            self[key] = set()
            return self[key]
        else:
            raise KeyError(str(key) + ' not in scheduler')
        
    def tick(self, dt):
        old = int(self.current)
        self.current += dt
        mod = self.mod
        for key in xrange(old, int(self.current)):
            key = (key % mod,)
            if key in self:
                for task in self[key]:
                    task()
                self[key].clear()
        self.current %= mod


class ResourceStore(object):
    def __init__(self, loader, prefix, suffix):
        self.loaded = {}
        self.loader = loader
        self.prefix = prefix
        self.suffix = suffix
        
    def __call__(self, resource):
        if resource not in self.loaded:
            resourceFile = os.path.join(self.prefix, resource + self.suffix)
            self.loaded[resource] = self.loader(resourceFile)
        return self.loaded[resource]
        
imageStore = ResourceStore(loadImage, 'data', '.png')
        
class World(object):
    def __init__(self, ground, dudes):
        self.floor = ground.outlineHeights
        self.groundBaddies = []
        self.aerialBaddies = []
        self.dudes = dudes
        self.collider = {}
        self.gravity = -0.5
    
        self.friction = 1 - 0.2
        self.air = 1 - 0.1
        self.lvlheight = 400
        broadcastSystem['tick'].add(self.tick)
        
    def groundMove(self, thing, dt, grav, fric, air):
        if not thing.resting:
            thing.yvel += thing.yacc * dt
            thing.yvel += grav
            resist = air
        else:
            thing.yvel = 0
            resist = fric
        thing.xvel += thing.xacc * dt
        if thing.yvel > 0:
            thing.yvel = min(thing.yvel, thing.yterm)
        else:
            thing.yvel = max(thing.yvel, -thing.yterm)
        thing.xvel *= resist
        absxvel = abs(thing.xvel)
        if absxvel > thing.xterm: 
            thing.xvel = copysign(thing.xterm, thing.xvel)
        elif absxvel < 0.1:
            thing.xvel = 0.0
        xpos = int(thing.x + thing.xvel*dt)
        xpos2 = xpos + thing.width
        maxAdj = abs(thing.x - xpos)
        climbHeight = thing.climbHeight if thing.resting else 0
        if thing.xvel > 0:
            while self.floor[xpos2] - climbHeight > thing.y and maxAdj:
                xpos2 -= 1
                xpos -= 1
                maxAdj -= 1
        elif thing.xvel < 0:
            while self.floor[xpos] - climbHeight > thing.y and maxAdj:
                xpos += 1
                xpos2 += 1
                maxAdj -=1
        thing.x = xpos
        floory = max(self.floor[xpos], self.floor[xpos2])
        thing.y = int(max(thing.y + thing.yvel*dt, floory))
        thing.resting = thing.y == floory
            
        thing.camera.setCameraPosition((thing.x, self.lvlheight-thing.y-thing.height))
        if thing.y == 0:
            thing.kill()
        
    def tick(self, dt):
        gravity = self.gravity
        gravDist = self.gravity * dt
        fric = self.friction
        air = self.air
        self.collider.clear()
        groundmove = self.groundMove
        dels = []
        for baddie in self.groundBaddies:
            groundmove(baddie, dt, gravDist, fric, air)
        for dude in self.dudes:
            groundmove(dude, dt, gravDist, fric, air)
            if dude.dead:
                dels.append(dude)
        for dude in dels:
            self.dudes.remove(dude)

class GroundMover(object):
    """Must have the following attributes:
       x, y, xvel, yvel, xacc, yacc
       And function setPosition
    """
    
    
class CameraManager(object):
    def __init__(self, following, cameras, zoom, windowSize, limits={}):
        self.cameras = cameras
        self.following = following
        for camera in self.cameras:
            camera.setCameraScale(zoom)
        self.zoom = float(zoom)
        self.windowSize = windowSize
        self.middleArea = (self.windowSize[0]*0.2, self.windowSize[1]*0.3,
                           self.windowSize[0]*0.8, self.windowSize[1]*0.7)
        self.offsetY = -self.windowSize[1]*(zoom - 1)
        self.current = {}
        self.current['x'] = 0
        self.current['y'] = 0
        self.limits = limits
        for limit, default in (('x', (0, 1000*zoom)), ('y', (0, 1000*zoom))): 
            if limit not in self.limits:
                self.limits[limit] = default
        broadcastSystem['tick'].add(self.tick)
        def lt(x, y):
            return x < y
        def gt(x, y):
            return x > y
        self.cameraMoveConditions = (
            (lt, 'y', self.middleArea[1], -1), (gt, 'y', self.middleArea[3], 1),
            (lt, 'x', self.middleArea[0], -1), (gt, 'x', self.middleArea[2], 1))
            
    def tick(self, dt):
        followx, followy = self.following.x, self.following.y
        display = {}
        display['x'] = (followx * self.zoom) - self.current['x']
        display['y'] = (followy * self.zoom) - self.current['y']
        for test, direction, boundVal, mult in self.cameraMoveConditions:
            if test(display[direction], boundVal):
                delta = abs(display[direction] - boundVal) 
                self.current[direction] += (delta * mult)
                lowerlimit, upperlimit = self.limits[direction]
                if self.current[direction] < lowerlimit:
                    self.current[direction] = lowerlimit
                elif self.current[direction] > upperlimit:
                    self.current[direction] = upperlimit
                self.dirty = True
        if self.dirty:
            self.dirty = False
            newPos = (-self.current['x'], self.current['y']+self.offsetY)
            for camera in self.cameras:
                camera.setCameraPosition(newPos)
        
    
class Dude(GroundMover):
    def __init__(self, groundCamera, scheduler):
        images = [imageStore(x) 
                  for x in ['leg', 'leg', 'body', 'arm', 'arm', 'head']]
        self.sprites = makeAdjacentSprites(images, 0)
        self.health = 100
        self.dead = False
        leg1, leg2, body, arm1, arm2, head = self.sprites
        body.setPosition((0, 5))
        leg1.setPosition((0, 10))
        leg2.setPosition((2, 10))
        arm1.setPosition((1, 7))
        arm2.setPosition((1, 7))
        self.climbHeight = 4
        self.xterm, self.yterm = 0.2, 0.3
        self.xacc, self.yacc = 0, 0
        self.xpacc, self.xnacc = 0, 0
        self.resting = False
        self.camera = TwoDVertexCamera()
        self.x, self.y, self.xvel, self.yvel = 20, 100, 0, 0
        self.width = 3
        self.height = 15
        self.jumping = False
        for sprite in self.sprites:
            sprite.setCameras([self.camera, groundCamera])
            sprite.setFilter(0)
        broadcastSystem[K_UP].add(self.jump)
        broadcastSystem[K_LEFT].add(self.moveLeft)
        broadcastSystem[K_RIGHT].add(self.moveRight)
        self.scheduler = scheduler
        self.jumpCancel = ScheduledTask(self.cancelJump)
        
    def kill(self):
        broadcastSystem[K_UP].remove(self.jump)
        broadcastSystem[K_LEFT].remove(self.moveLeft)
        broadcastSystem[K_RIGHT].remove(self.moveRight)
        del self.sprites
        self.dead = True
        
    def moveRight(self, keydown):
        self.xpacc = 0.2 if keydown else 0
        self.xacc = self.xnacc + self.xpacc
        
    def moveLeft(self, keydown):
        self.xnacc = -0.2 if keydown else 0
        self.xacc = self.xnacc + self.xpacc
        
    def cancelJump(self):
        if self.jumping:
            self.yacc = min(0, self.yacc)
            self.jumping = False
        
    def jump(self, keydown):
        if keydown:
            if self.resting:
                self.yacc = 1
                self.scheduler[140].add(self.jumpCancel)
                self.resting = False
                self.jumping = True
        else:
            self.cancelJump()
            
    def setPosition(self, newpos):
        self.camera.setPosition(newpos)
        

class HeightDict(dict):
    def __missing__(self, key):
        return 0
        
        
class Ground(object):
    def __init__(self, lvlheight, camera, fragCamera):
        self.sprites = []
        self.heights = HeightDict()
        self.outlineHeights = HeightDict()
        
        land = ['flat', 'uphill', 'downhill']
        absence = ['hole']
        vert = ['wall', 'cliff']
        all = land + absence + vert
        features = {'flat': (lambda x: (x, x), 100, all),
                    'uphill': (lambda x: (min(40, x+1), min(40, x+1)), 20, all),
                    'downhill': (lambda x: (max(1, x-1), max(1, x-1)), 20, all),
                    'hole': (lambda x: (0, x), 4, land),
                    'wall': (lambda x: (min(40, x+4), min(40, x+4)), 1, land),
                    'cliff': (lambda x: (max(1, x-8), max(1, x-8)), 1, land),
                    }
                    
        currentFeature = 'flat'
        featureChangeProb = 0.1
        currentHeight = 20
        currentPos = 0
        featureFunc, featureLenRemaining, nextFeatures = \
            features[currentFeature]
        
        for x in xrange(10):
            # generate flat at beginning
            self.heights[currentPos], currentHeight = featureFunc(currentHeight)
            currentPos += 1
            featureLenRemaining -= 1
            
        for x in xrange(500):
            if random.random() < featureChangeProb or not featureLenRemaining:
                currentFeature = random.choice(nextFeatures)
                featureFunc, featureLenRemaining, nextFeatures = \
                    features[currentFeature]
                nextFeatures = nextFeatures[:]
                if currentFeature in nextFeatures:
                    nextFeatures.remove(currentFeature)
            self.heights[currentPos], currentHeight = featureFunc(currentHeight)
            featureLenRemaining -= 1 
            currentPos += 1
        
        for x in xrange(min(self.heights), max(self.heights)):
            height = self.heights[x]
            xcoord = 8 * x
            ycoord = lvlheight - (8 * height)
            if self.heights[x-1] < height and self.heights[x+1] < height:
                topimage = imageStore('tipblock')
                for z in xrange(4):
                    self.outlineHeights[xcoord + z] = 8 * (height - 1) + z
                for z in xrange(4):
                    self.outlineHeights[xcoord + 4 + z] = 8 * height - (z + 4)
            elif self.heights[x-1] < height:
                topimage = imageStore('slopeblock')
                for z in xrange(8):
                    self.outlineHeights[xcoord + z] = 8 * (height - 1) + z
            elif self.heights[x+1] < height:
                topimage = imageStore('slopeblock2')
                for z in xrange(8):
                    self.outlineHeights[xcoord + z] = 8 * height - z
            else:
                topimage = imageStore('topblock')
                for z in xrange(8):
                    self.outlineHeights[xcoord + z] = 8 * height
            baseimage = imageStore('baseblock')
            
            if height != 0:
                self.sprites.append(Sprite(position=(xcoord, ycoord), 
                    image=topimage, offsets=topLeftOffsets, offsetsRel=True,
                    layer=-1, cameras=[camera], fragCameras=[fragCamera]))
                
                times = int((lvlheight - ycoord) / 8)
                ycoord += 8
                self.sprites.append(Sprite(position=(xcoord, ycoord),
                        image=baseimage, offsets=topLeftOffsets, offsetsRel=True,
                        tiling = (1, times),
                        layer=-1, cameras=[camera], fragCameras=[fragCamera]))
                            

class HealthDisplay(object):
    def __init__(self, track, zoom):
        self.tracked = track
        self.currentDisplay = self.tracked.health
        self.images = dict([(str(x), imageStore(str(x))) for x in xrange(10)])
        self.images['%'] = imageStore('percent')
        self.camera = TwoDVertexCamera()
        self.camera.setCameraScale(zoom)
        cameras = [self.camera]
        self.digit0 = Sprite(image=self.images['1'], position=(0, 0), layer=-1,
                             cameras=cameras)
        self.digit1 = Sprite(image=self.images['0'], position=(9, 0), layer=-1,
                             cameras=cameras)
        self.digit2 = Sprite(image=self.images['0'], position=(18, 0), layer=-1,
                             cameras=cameras)
        self.percent = Sprite(image=self.images['%'], position=(27, 0), layer=-1,
                             cameras=cameras)
        broadcastSystem['tick'].add(self.tick)
    
    def tick(self, dt):
        if self.tracked.health != self.currentDisplay:
            self.currentDisplay = self.tracked.health
            strdisplay = str(self.currentDisplay).rjust(3, '0')
            for x, sprite in ((0, self.digit0), (1, self.digit1), 
                             (2, self.digit2)):
                sprite.setImage(self.images[strdisplay[x]])
        
    
class Baddie(object):
    pass
    
class Game(object):
    def __init__(self):
        self.windowSize = (800, 400)

    def main(self, broadcastSystem=broadcastSystem):
        GLCommon.initEnvironment(self.windowSize)
        SpriteCommon.setDefaultFilter('nearest')
        clock = pygame.time.Clock()
        groundEnemyCam = TwoDVertexCamera()
        fadingCam = TwoDFragmentCamera()
        zoom = 4.0
        scheduler = Scheduler(0, 5000)
        x = Ground(self.windowSize[1], groundEnemyCam, fadingCam)
        d = Dude(groundEnemyCam, scheduler)
        camManager = CameraManager(d, [groundEnemyCam], zoom, self.windowSize)
        health = HealthDisplay(d, zoom)
        world = World(x, [d])
        going = True
        while going:
            t = clock.tick(60)
            broadcastSystem['tick'].broadcast(t)
            for event in pygame.event.get():
                if event.type == QUIT:
                    going = False
                elif event.type == KEYDOWN and event.key == K_ESCAPE:
                    going = False
                elif event.type in (KEYDOWN, KEYUP):
                    if event.key in broadcastSystem:
                        down = event.type == KEYDOWN
                        broadcastSystem[event.key].broadcast(down)
            pygl3display.GLCommon.doTasks()
            if d.dead:
                going = False
        sys.exit(0)

def main():
    x = Game()
    x.main()
    
if __name__ == '__main__':
    if 'debug' in sys.argv:
        DEBUG = True
    if 'profile' in sys.argv:
        import cProfile
        cProfile.run('main()')
    else:
        main()

    
