"""
PyGL3Display
(C) David Griffin (2010-present)
"Habilain" is a pseudonym used by David Griffin in a number of places.

A high performance Pygame OpenGL library, capable of handling 1000s of sprites

Requirements: Python, Pygame, PyOpenGL, Open GL 2.0 graphics hardware

Licensed under GPLv2 or GPLv3. You should have received a copy of both of these
licenses in the files GPLv2.txt and GPLv3.txt. If not these licenses can be
found at http://www.gnu.org/licenses/gpl-2.0.html (GPLv2) and
http://www.gnu.org/licenses/gpl-3.0.html (GPLv3). By using this file you agree
to be bound by at least one of these licenses.

Other licenses, including commercial/proprietary licenses, can be arranged by 
contacting me via e-mail at habilain@gmail.com. If you do not have in physical 
writing an alternative license agreement from me, you must use one of the 
licenses specified above when releasing work which makes use of this library.

pygl3display.gl.cffigl.cffigl

Provides the experimental and likely highly buggy CFFI binding"""

from itertools import izip
from gl_raw import RawBinding, glInts, glfwInts, glewInts, gl3Ints
from functools import wraps
glc = RawBinding(glInts + gl3Ints+ glfwInts + glewInts)
import os.path

ACCEPT_PYTHON_DATA = True # If False, Only CData objects are allowed as input.
DEBUG = True # Call GLError a lot
docStringTemplates = {}
docStringTemplates['glGet'] = \
"""glGet%s: if len(glGet%sv(pname)) == 1, return that %s.
Otherwise, an error.
Args:
 pname: Parameter to query"""
docStringTemplates['glGetv'] = \
"""glGet%sv: returns a tuple of %s parameters 
associated with pname
Args:
 pname: Parameter to query"""

docStrings = {}
for dataType in ['Integer', 'Boolean', 'Float', 'Double']:
    docStrings['glGet%s' % dataType] = \
        docStringTemplates['glGet'] % ((dataType,) * 3)
    docStrings['glGet%sv' % dataType] = \
        docStringTemplates['glGetv'] % ((dataType,) * 2)
 
docStringTemplates['specBufferData'] = \
"""As glBufferData, but assumes that the data is of type %s.
This allows native Python data types to be automatically converted.
Also does not require the size of the data to be inputted; this is
automatically calculated."""
docStringTemplates['specBufferSubData'] = \
"""As glBufferSubData, but assumes that the data is of type %s.
This allows native Python data types to be automatically converted.
Also does not require the size of the data to be inputted; this is
automatically calculated. Also, offset is measured in the given
units rather than bytes."""

class GLMetaClass(type):
    def __init__(cls, name, bases, dictionary):
        super(GLMetaClass, cls).__init__(name, bases, dictionary)
        cls.dataTypesDict = {'b': ('GLbyte', 1, glc.GL_BYTE),
                              'h': ('GLshort', 2, glc.GL_SHORT),
                              'i': ('GLint', 4, glc.GL_INT),
                              'B': ('GLubyte', 1, glc.GL_UNSIGNED_BYTE),
                              'H': ('GLushort', 2, glc.GL_UNSIGNED_SHORT),
                              'I': ('GLuint', 4, glc.GL_UNSIGNED_INT),
                              'f': ('GLfloat', 4, glc.GL_FLOAT),
                              'd': ('GLdouble', 8, glc.GL_DOUBLE),
                              'bool': ('GLboolean', 1, glc.GL_UNSIGNED_BYTE)}
        cls.convDict = cls.makeConvDict(cls.dataTypesDict.keys())
        cls.componentSizes = {glc.GL_RED: 1, glc.GL_GREEN: 1, glc.GL_BLUE: 1,
                              glc.GL_RG: 2, glc.GL_RGB: 3, glc.GL_RGBA: 4,
                              glc.GL_BGR: 3, glc.GL_BGRA: 4,}
        for funcStem, datatype in (('glGetInteger', 'i'), ('glGetBoolean', 'bool'),
                                   ('glGetFloat', 'f'), ('glGetDouble', 'd')):
            cls.makeAllGetters(funcStem, datatype)
        cls.makeGLGens()
        cls.makeShaderGetters()
        cls.makeUniformSetters()
        for dataType, suffix in (('f', 'f'), ('bool', 'bool'), ('i', 'i'), 
                                 ('I', 'ui'), ('b', 'b'), ('B', 'ub')):
            cls.makeBufferData(dataType, suffix)
        for name, csig in (('glfwSetKeyCallback', 'void(int, int)'), 
                           ('glfwSetCharCallback', 'void(int)'),
                           ('glfwSetMouseButtonCallback', 'void(int, int)'),
                           ('glfwSetMouseWheelCallback', 'void(int)'),
                           ('glfwSetMouseButtonCallback', 'void(int, int)'),):
            cls.makeCallbackFunc(name, csig)
        cls.makeIntTypeSensingFuncs()
        cls.makeListAcceptors()
    
    def makeConvDict(cls, dataTypes):
        convDict = {}
        for dataType in dataTypes:
            ctype, size, glname = cls.dataTypesDict[dataType] 
            convDict[glname] = '%s[]' % ctype
        return convDict
    
    def acceptPyData2(cls, dataTypes, argName='data', argPos=-1, 
                      dataTypeArg='type', dataTypeArgPos=-2, static=False):
        """Decorator which converts an argument to a cData object by
        inspecting a parameter which determines the type to convert to.
        If the function is 'static', then it does not receive the self
        argument that the wrapped function will receive"""
        
        convDict = cls.makeConvDict(dataTypes)
                
        def wrap(func):
            global ACCEPT_PYTHON_DATA
            if ACCEPT_PYTHON_DATA:
                def convertData(self, cArrayType, data):
                    if not isinstance(data, type(self._ffi.NULL)):
                        data = self._ffi.new(cArrayType, data)
                    return data
                
                def wrapped_func(self, *args, **kwargs):
                    if dataTypeArg in kwargs:
                        glDataType = kwargs[dataTypeArg]
                    else:
                        glDataType = args[dataTypeArgPos]
                    cArrayType = convDict[glDataType]
                    if argName in kwargs:
                        kwargs[argName] = convertData(self, cArrayType, kwargs[argName])
                    else:
                        argPosAbs = argPos % len(args)
                        data = args[argPosAbs]                                        
                        args = args[:argPosAbs] + (convertData(self, cArrayType, data), ) + args[(argPosAbs+1):]
                    if static:
                        return func(*args, **kwargs)
                    else:
                        return func(self, *args, **kwargs)
                return wrapped_func
            else:
                return func
        return wrap
                    
    def acceptPyData(cls, dataType, argName='data', argPos=-1):
        """Decorator which converts an argument to a cData object"""
        def wrap(func):
            global ACCEPT_PYTHON_DATA
            if ACCEPT_PYTHON_DATA:
                arrayType = '%s[]' % dataType
                def convertData(self, data):
                    if not isinstance(data, type(self._ffi.NULL)):
                        data = self._ffi.new(arrayType, data)
                    return data
                
                def wrapped_func(self, *args, **kwargs):
                    if argName in kwargs:
                        kwargs[argName] = convertData(self, kwargs[argName])
                    else:
                        argPosAbs = argPos % len(args)
                        data = args[argPosAbs]                                        
                        args = args[:argPosAbs] + (convertData(self, data), ) + args[(argPosAbs+1):]    
                    return func(self, *args, **kwargs)
                
                return wrapped_func
            else:
                return func
        return wrap

    def finaliseFunc(cls, name, doc, func):
        """Finalises a metaprogrammed function by assigning its name, docstring
        and assigning it to the class"""
        func.__name__ = name
        func.__doc__ = doc
        setattr(cls, name, func)
        
    def makeAllGetters(cls, funcStem, dataType):
        cdataType = cls.dataTypesDict[dataType][0]
        arrayType = '%s[]' % cdataType
        cls.makeSingleGetter(funcStem, arrayType)
        cls.makeVectorGetter(funcStem, arrayType)
        
    def makeShaderGetters(cls):
        cls.makeParamSingleGetter('glGetShaderi', 'GLint[]')
        cls.makeParamSingleGetter('glGetProgrami', 'GLint[]')
        cls.makeParamVectorGetter('glGetShaderi', 'GLint[]')
        cls.makeParamVectorGetter('glGetProgrami', 'GLint[]')
        cls.makeTwoParamSingleGetter('glGetTexLevelParameterf', 'GLfloat[]')
        cls.makeTwoParamSingleGetter('glGetTexLevelParameteri', 'GLint[]')
        cls.makeTwoParamVectorGetter('glGetTexLevelParameterf', 'GLfloat[]')
        cls.makeTwoParamVectorGetter('glGetTexLevelParameteri', 'GLint[]')
        cls.makeInfoLogGetter('glGetShaderInfoLog', 'glGetShaderi', glc.GL_INFO_LOG_LENGTH)
        cls.makeInfoLogGetter('glGetProgramInfoLog', 'glGetProgrami', glc.GL_INFO_LOG_LENGTH)
        
    def makeUniformSetters(cls):
        for glFunc in ('glUniform1fv', 'glUniform2fv', 'glUniform3fv', 
                       'glUniform4fv'):
            cls.makeUniformSetter(glFunc, 'GLfloat')
        for glFunc in ('glUniform1iv', 'glUniform2iv', 'glUniform3iv', 
                       'glUniform4iv'):
            cls.makeUniformSetter(glFunc, 'GLint')
        for glFunc in ('glUniformMatrix2fv', 'glUniformMatrix3fv', 
                       'glUniformMatrix4fv'):
            cls.makeUniformMatrixSetter(glFunc, 'GLfloat')
                
    def makeUniformMatrixSetter(cls, glFuncName, dataType):
        func = getattr(glc, glFuncName)
        @cls.acceptPyData(dataType)
        def uniformMatrixSetter(self, location, count, transpose, data):
            func(location, count, transpose, data)
        cls.finaliseFunc(glFuncName, '', uniformMatrixSetter)

    def makeUniformSetter(cls, glFuncName, dataType):
        func = getattr(glc, glFuncName)
        @cls.acceptPyData(dataType)
        def uniformSetter(self, location, count, data):
            func(location, count, data)
        cls.finaliseFunc(glFuncName, '', uniformSetter)
        
    def makeGLGens(cls):
        def makeFuncs(singular):
            plural = '%ss' % singular
            func = getattr(glc, plural)
            def singularWrap(self):
                x = self._ffi.new(pointerType, 1)
                func(1, x)
                return x[0]
            def pluralWrap(self, number):
                x = self._ffi.new(pointerType, number)
                func(number, x)
                return tuple(x)
            cls.finaliseFunc(singular, '', singularWrap)
            cls.finaliseFunc(plural, '', pluralWrap) 
        dataType = 'GLuint'
        pointerType = '%s[]' % dataType
        for singular in ('glGenBuffer', 'glGenVertexArray', 'glGenTexture',
                         'glGenFramebuffer'):
            makeFuncs(singular)
        
    def makeSingleGetter(cls, funcStem, dataType):
        glFunc = funcStem + 'v'
        funcName = funcStem
        func = getattr(glc, glFunc)
        def singleGetter(self, pname):
            if pname not in self._specPNames:
                x = self._ffi.new(dataType, 1)
                func(pname, x)
                return x[0]
            else:
                raise Exception("%s returns a tuple of values" % self._pnames[pname])
        cls.finaliseFunc(funcName, docStrings[funcName], singleGetter)
    
    def makeVectorGetter(cls, funcStem, dataType):
        glFunc = funcStem + 'v'
        funcName = glFunc
        func = getattr(glc, glFunc)
        def vectorGetter(self, pname):
            x = self._ffi.new(dataType, self._specPNames.get(pname, 1))
            func(pname, x)
            return tuple(x)
        cls.finaliseFunc(funcName, docStrings[funcName], vectorGetter)
        
    def makeParamSingleGetter(cls, funcStem, dataType):
        glFunc = funcStem + 'v'
        funcName = funcStem
        func = getattr(glc, glFunc)
        def paramSingleGetter(self, param, pname):
            if pname not in self._specPNames:
                x = self._ffi.new(dataType, 1)
                func(param, pname, x)
                return x[0]
            else:
                raise Exception("%s returns a tuple of values" % self._pnames[pname])
        cls.finaliseFunc(funcName, '', paramSingleGetter)

    def makeParamVectorGetter(cls, funcStem, dataType):
        glFunc = funcStem + 'v'
        funcName = glFunc
        func = getattr(glc, glFunc)
        def paramVectorGetter(self, param, pname):
            x = self._ffi.new(dataType, self._specPNames.get(pname, 1))
            func(param, pname, x)
            return tuple(x)
        cls.finaliseFunc(funcName, '', paramVectorGetter)
        
    def makeTwoParamSingleGetter(cls, funcStem, dataType):
        glFunc = funcStem + 'v'
        funcName = funcStem
        func = getattr(glc, glFunc)
        def paramSingleGetter(self, param1, param2, pname):
            if pname not in self._specPNames:
                x = self._ffi.new(dataType, 1)
                func(param1, param2, pname, x)
                return x[0]
            else:
                raise Exception("%s returns a tuple of values" % self._pnames[pname])
        cls.finaliseFunc(funcName, '', paramSingleGetter)
        
    def makeTwoParamVectorGetter(cls, funcStem, dataType):
        glFunc = funcStem + 'v'
        funcName = glFunc
        func = getattr(glc, glFunc)
        def paramVectorGetter(self, param1, param2, pname):
            x = self._ffi.new(dataType, self._specPNames.get(pname, 1))
            func(param1, param2, pname, x)
            return tuple(x)
        cls.finaliseFunc(funcName, '', paramVectorGetter)
        
    def makeInfoLogGetter(cls, glFunc, queryFuncName, pname):
        funcName = glFunc
        func = getattr(glc, glFunc)
        def infoLogGetter(self, param, length=None):
            if length is None:
                queryFunc = getattr(self, queryFuncName)
                length = queryFunc(param, pname)
            x = self._ffi.new('GLchar[]', length)
            func(param, length, self._ffi.NULL, x)
            return self._ffi.string(x)
        cls.finaliseFunc(funcName, '', infoLogGetter)
        
    def makeIntTypeSensingFuncs(cls):
        dataTypes = 'bBhHiIf'
        convertor = cls.acceptPyData2(dataTypes, 'data', -1, 'type', -2, static=True)
        for funcName in ('glTexImage1D', 'glTexImage2D', 'glTexImage3D',
                         'glTexSubImage1D', 'glTexSubImage2D', 
                         'glTexSubImage3D', 'glDrawElements'):
            wrappeFunc = convertor(getattr(glc, funcName))
            cls.finaliseFunc(funcName, '', wrappeFunc)
        
    def makeBufferData(cls, dataType, suffix):
        cdataType, dataSize = cls.dataTypesDict[dataType][0:2]
        glBufferData = glc.glBufferData
        @cls.acceptPyData(cdataType, 'data', 1)
        def bufferData(self, target, data, usage):
            return glBufferData(target, self._ffi.sizeof(data), data, usage)
        docs = docStringTemplates['specBufferData'] % dataType
        cls.finaliseFunc('glBufferData%s' % suffix, docs, bufferData)
        glBufferSubData = glc.glBufferSubData
        @cls.acceptPyData(cdataType, 'data', 2)
        def bufferSubData(self, target, offset, data):
            return glBufferSubData(target, dataSize*offset, self._ffi.sizeof(data), data)
        docs = docStringTemplates['specBufferSubData'] % dataType
        cls.finaliseFunc('glBufferSubData%s' % suffix, docs, bufferSubData)
            
    def makeListAcceptors(cls):
        def makeFuncs(singular):
            plural = '%ss' % singular
            glFunc = getattr(glc, plural)
            @cls.acceptPyData('GLuint', 'param', 0)
            def pluralWrap(self, param):
                glFunc(len(param), param)
            def singularWrap(self, param):
                x = self._ffi.new('GLuint[]', 1)
                x[0] = param
                glFunc(1, x)
            cls.finaliseFunc(plural, '', pluralWrap)
            cls.finaliseFunc(singular, '', singularWrap)        

        for singular in ('glDeleteTexture', 'glDrawBuffer'):
            makeFuncs(singular)
            
    def makeCallbackFunc(cls, name, csig):
        glfwcallbackfunc = getattr(glc, name)
        def callback(self, function):
            funcptr = self._ffi.callback(csig, function)
            setattr(self, '%sCallback' % name, funcptr)
            glfwcallbackfunc(funcptr)
        cls.finaliseFunc(name, '', callback) 
            
class CStringList(object):
    def __init__(self, ffi, strings):
        self.cstrings = [ffi.new('char[]', string) for string in strings]
        self.cstringList = ffi.new('char*[]', self.cstrings)
        
    def __len__(self):
        return len(self.cstrings)
    
    def data(self):
        return self.cstringList         

class CArray(object):
    def __init__(self, cdata):
        self.cdata = cdata
        
    def __getattr__(self, key):
        return getattr(self.cdata, key)
        
    def __getitem__(self, indx):
        if isinstance(indx, slice):
            step = indx.step if indx.step is not None else 1
            return [self.cdata.__getitem__(sindx) for sindx in 
                    xrange(indx.start, indx.stop, step)]
        else:
            return self.cdata.__getitem__(indx)
        
    def __setitem__(self, indx, val):
        if isinstance(indx, slice):
            step = indx.step if indx.step is not None else 1
            for sindx, sval in izip(xrange(indx.start, indx.stop, step), val):
                self.cdata.__setitem__(sindx, sval)
        else:
            self.cdata.__setitem__(indx, val)
            
# TODO: glGenFramebuffers, glFramebufferTexture2D, glCheckFramebufferStatus, 
# glBindFramebuffer, GL_FRAMEBUFFER, GL_FRAMEBUFFER_COMPLETE are all OpenGL 3.0 only.
# I need a method to get the EXT versions of these as appropriate, as calling
# any of these will be ugly.
# NOTE2: Those functions are not 3.0 only. Rather there's an ARB extension that
# may be defined, and if it is they work on 2.1. Will need a fallback, but for now
# I don't need to do anything here.

class GL(object):
    __metaclass__ = GLMetaClass
    
    def __init__(self):
        self.__glc = glc
        self._ffi = glc.ffi 
        self._pnames = {} # Dictionary of gl integer names to human readable names
        self.inited = False
        for name in glc.constNames:
            self._pnames[glc[name]] = name
        for name in self.__glc.__dict__:
            if not hasattr(type(self), name):
                setattr(self, name, getattr(self.__glc, name))
        self._specPNames = {} # Dictionary of values that return multiple values - not set up yet.
        
        self.cdata = {'c': lambda x: self._ffi.new('GLchar[]', x),
         'b': lambda x: self._ffi.new('GLbyte[]', x),
         'B': lambda x: self._ffi.new('GLubyte[]', x),
         'h': lambda x: self._ffi.new('GLshort[]', x),
         'H': lambda x: self._ffi.new('GLushort[]', x),
         'i': lambda x: self._ffi.new('GLint[]', x),
         'I': lambda x: self._ffi.new('GLuint[]', x),
         #'l': lambda x: self._ffi.new('GL[]', x),
         #'L': lambda x: self._ffi.new('GL[]', x),
         'f': lambda x: self._ffi.new('GLfloat[]', x),
         'd': lambda x: self._ffi.new('GLdouble[]', x),
         } 
        self.cdataSizes = {'c': 1, 'b': 1, 'B': 1, 'h': 2, 'H': 2, 
              'i': 4, 'I': 4, 'l': 8, 'L': 8, 'f': 4, 'd': 8}
        self.wrappers = {} # Wrapper functions
        
    def _makeShellBindings(self):
        shellBindings = ['"""shellbindings.py', 
                         'An automatically generated script designed to keep IDEs and such like happy.',
                         'Seriously, without it you get an awful lot of errors"""']
        for name in dir(self):
            if not name.startswith('_'):
                shellBindings.append('%s = None' % name)
        directory = os.path.dirname(__file__)
        with open(os.path.join(directory, 'shellbindings.py'), 'w') as f:
            f.write('\n'.join(shellBindings))
            
    def _makeBindings(self):
        bindings = ['"""bindigs.py', 
                    'An automatically generated binding of attributes from the cffigl GL object',
                    'onto something that can be imported. This will need to be rebuilt whenever',
                    'the cffigl binding or GL object change. This is done to keep IDEs and Pypy',
                    'happy - they don\'t like other methods."""',
                    '',
                    'from pygl3display.gl.cffigl.cffigl import GL',
                    '__binding = GL()']
        for name in self.__dict__.keys() + type(self).__dict__.keys():
            if not name.startswith('_'):
                bindings.append('%s = getattr(__binding, """%s\""")' % (name, name))
        directory = os.path.dirname(__file__)
        with open(os.path.join(directory, 'bindings.py'), 'w') as f:
            f.write('\n'.join(bindings))
    
    def byref(self, data, offset=0):
        # Assumes data is a pointer type
        return data + offset
    
    def sizeOf(self, data):
        return self._ffi.sizeof(data)
    
    def makeCStrings(self, strings):
        return CStringList(self._ffi, strings)
        
    def new(self, *args, **kwargs):
        return self._ffi.new(*args, **kwargs)
    
    def toByteString(self, cdata):
        return self._ffi.buffer(cdata)[:]
        
    def offsetPointer(self, offset):
        return self._ffi.cast("void *", offset)
                
    def glGetString(self, pname):
        return self._ffi.string(self.__glc.glGetString(pname))
    
    def glGetStringi(self, pname, index):
        return self._ffi.string(self.__glc.glGetStringi(pname, index))
    
    def glGetError(self, lastFunc=''):
        errCode = self.__glc.glGetError()
        if errCode != self.__glc.GL_FALSE:
            raise Exception('GL Error %s:%s' % (lastFunc, self._pnames[errCode])) 
        
    def glGetTexImage(self, target, level, formt, dtype):
        width = self.glGetTexLevelParameteri(target, level, self.GL_TEXTURE_WIDTH)
        height = self.glGetTexLevelParameteri(target, level, self.GL_TEXTURE_HEIGHT)
        arrayType = type(self).convDict[dtype]
        noComponents = type(self).componentSizes[formt]
        cdata = self.new(arrayType, width * height * noComponents)
        self.__glc.glGetTexImage(target, level, formt, dtype, cdata)
        return cdata
    
    def glReadPixels(self, x, y, width, height, formt, dtype):
        arrayType = type(self).convDict[dtype]
        noComponents = type(self).componentSizes[formt]
        cdata = self.new(arrayType, width * height * noComponents)
        self.__glc.glReadPixels(x, y, width, height, formt, dtype, cdata)
        return cdata
    
    def glShaderSource(self, shader, source):
        if isinstance(source, str):
            source = [source]
        cSource = self.makeCStrings(source)
        self.__glc.glShaderSource(shader, len(cSource), cSource.data(), self.__glc.ffi.NULL)
        
    def glewInit(self):
        self.__glc.glewInit()
        self.inited = True
        
    if DEBUG:
        def __getattribute__(self, key):
            thing = super(GL, self).__getattribute__(key)
            if key != 'glGetError' and callable(thing):
                # Under pypy, cannot use wraps. But wraps is kinda necessary for making bindings. That or disable debug.
                #@wraps(thing)
                if key not in self.wrappers:
                    def wrap(*args, **kwargs):
                        if not self.inited and key != 'glewInit' and not key.startswith('glfw') and not key.startswith('_'):
                            raise Exception('Called GL function %s before glewInit' % str(thing))
                        ret = thing(*args, **kwargs)
                        self.glGetError(key)
                        return ret
                    self.wrappers[key] = wrap
                return self.wrappers[key]
            else:
                return thing

        