"""
PyGL3Display
(C) David Griffin (2010-present)
"Habilain" is a pseudonym used by David Griffin in a number of places.

A high performance Pygame OpenGL library, capable of handling 1000s of sprites

Requirements: Python, Pygame, PyOpenGL, Open GL 2.0 graphics hardware

Licensed under GPLv2 or GPLv3. You should have received a copy of both of these
licenses in the files GPLv2.txt and GPLv3.txt. If not these licenses can be
found at http://www.gnu.org/licenses/gpl-2.0.html (GPLv2) and
http://www.gnu.org/licenses/gpl-3.0.html (GPLv3). By using this file you agree
to be bound by at least one of these licenses.

Other licenses, including commercial/proprietary licenses, can be arranged by 
contacting me via e-mail at habilain@gmail.com. If you do not have in 
writing an alternative license agreement from me, you must use one of the 
licenses specified above when releasing work which makes use of this library.

pygl3display.noneweakkey

Provides a weak key dictionary extended so that it can also store a None value
"""

import weakref

class NoneWeakKeyDict(weakref.WeakKeyDictionary):
    """A minor modification of a weak key dictioanry that supports
    storing the none value as a key. This is necessary as "None" is a
    sensible value for having no render target, but WeakRefs are necessary
    to avoid memory leaks with actual render targets"""
    
    def __contains__(self, key):
        """Check if we have a None attribute"""
        if key is None:
            return hasattr(self, 'noneval')
        else:
            return weakref.WeakKeyDictionary.__contains__(self, key)
            
    def __getitem__(self, key):
        """Return the None value if necessary (note: exception raised for None
        not in dictionary is attribute error rather than key error - this is
        OK in PyGL3Display but poor practice in general)"""
        if key is None:
            return self.noneval
        else:
            return weakref.WeakKeyDictionary.__getitem__(self, key)
            
    def __setitem__(self, key, val):
        """Allow setting of a none key"""
        if key is None:
            self.noneval = val
        else:
            weakref.WeakKeyDictionary.__setitem__(self, key, val)
