"""
PyGL3Display
(C) David Griffin (2010-present)
"Habilain" is a pseudonym used by David Griffin in a number of places.

A high performance Pygame OpenGL library, capable of handling 1000s of sprites

Requirements: Python, Pygame, PyOpenGL, Open GL 2.0 graphics hardware

Licensed under GPLv2 or GPLv3. You should have received a copy of both of these
licenses in the files GPLv2.txt and GPLv3.txt. If not these licenses can be
found at http://www.gnu.org/licenses/gpl-2.0.html (GPLv2) and
http://www.gnu.org/licenses/gpl-3.0.html (GPLv3). By using this file you agree
to be bound by at least one of these licenses.

Other licenses, including commercial/proprietary licenses, can be arranged by 
contacting me via e-mail at habilain@gmail.com. If you do not have in 
writing an alternative license agreement from me, you must use one of the 
licenses specified above when releasing work which makes use of this library.

pygl3display.onedtexture

Provides OneDTexture, an abstraction of a 1d texture, as well as some support
classes
"""

from array import array
from pygl3display.texture import Texture, colorMasks
import struct
from pygl3display.gl import (glTexImage1D, glGetTexLevelParameteriv,
    glBindTexture, glTexSubImage1D, glGetIntegerv, GLError, 
    GL_PROXY_TEXTURE_1D, GL_TEXTURE_1D, GL_RGBA, GL_UNSIGNED_BYTE,
    GL_TEXTURE_WIDTH, GL_RGBA32F, cdata)
    
__all__ = ['OneDTextureCommon', 'OneDTexture']
    
class OneDTextureCommon(object):
    """The shared state for OneD Texture Objects"""
    maxTextureSize = 8192
    minTextureSize = 512
    
    @classmethod
    def findTextureSize(cls, minsize=None, maxsize=None):
        """Function to determine maximum atlas size presently possible; 
        can specify a range that this has to be in as well."""
        if minsize is None: 
            minsize = cls.minTextureSize
        if maxsize is None: 
            maxsize = min(glGetIntegerv(GL_MAX_TEXTURE_SIZE), 
                          cls.maxTextureSize)
        size = maxsize
        while not cls.testSize(size):
            size >>= 1
        while not cls.testSize2(size):
            size >>= 1
        if size <= minsize:
            raise MemoryError('Cannot create an atlas of minimum size')        
        return size
            
    @staticmethod
    def testSize(texsize):
        """Helper function to determine maximum atlas size
        Arguments:
         texsize : test if an atlas of texsizextexsize can be created"""
        try:
            glTexImage1D(GL_PROXY_TEXTURE_1D, 0, GL_RGBA, 
                         texsize, texsize, 0, GL_RGBA, GL_UNSIGNED_BYTE, None)
            return glGetTexLevelParameteriv(GL_PROXY_TEXTURE_1D, 
                        0, GL_TEXTURE_WIDTH)
        except GLError:
            return False
            
    @staticmethod
    def testSize2(texsize):
        """Helper function to determine maximum atlas size.
        Some (bad) GL drivers want some dummy data, it seems,
        so this version gives dummy data. It's in a seperate function because
        for compliant systems creating a ctypes array for each texsize would
        give a high performance overhead. With this, there is low overhead
        on a compliant system as it minimises the number of different types
        of call.
        Arguments:
         texsize : test if an atlas of texsize can be created"""
        try:
            glTexImage1D(GL_PROXY_TEXTURE_1D, 0, GL_RGBA, 
                         texsize, texsize, 0, GL_RGBA, GL_UNSIGNED_BYTE, 
                         cdata['f'](texsize))
            return glGetTexLevelParameteriv(GL_PROXY_TEXTURE_1D, 
                        0, GL_TEXTURE_WIDTH)
        except GLError:
            return False    
    
class OneDTexture(Texture):
    """A One D Hardware accelerated texture"""
    def __init__(self, size, data = False):
        """Initialise a texture of given size. Will raise errors if texture
        of given size cannot be created; use the functions in OneDTextureCommon
        to determine possible texture sizes
        Args:
            size: size of the twodtexture
            data: If True, use GL_RGBA32F rather than GL_RGBA - this removes 
                  clamping norally present in textures, but may result in
                  a performance hit. Only useful if using GPU shaders for
                  compute operations.
        """

        internalType = GL_RGBA32F if data else GL_RGBA
        initialiser =  lambda: glTexImage1D(GL_TEXTURE_1D, 0, internalType, 
                size, 0, GL_RGBA, GL_UNSIGNED_BYTE, cdata['f'](size))
        self.initTexture(size, data, GL_TEXTURE_1D, initialiser)
                                     
    def upload(self, pos, size, data, mask = GL_RGBA):
        """Upload to a area on the texture
        Arguments:
         pos: The x coordinates of the top left of the area
         size: The width of the area
         data: The data to upload; must be size*4 bytes long
         mask: A mask for the data to upload; defaults to RGBA
        """
        mask = colorMasks.get(mask, mask)
        glBindTexture(GL_TEXTURE_1D, self.surface)
        glTexSubImage1D(GL_TEXTURE_1D, 0, pos, size, 
                        mask, GL_UNSIGNED_BYTE, data)
                                
    def readPixelsToArrays(self):
        """Wrestles pixels out of numpy arrays into python arrays, but this
        is slow, and probably should be avoided"""
        npyarray = self.readPixels()
        stringres = npyarray.tostring()
        res = [array('f', struct.unpack('f' * self.size * 4, 
                stringres[(x * self.size * 16):((x+1) * self.size * 16)])) 
                for x in xrange(self.size)]
        return res
        

