"""
PyGL3Display
(C) David Griffin (2010-present)
"Habilain" is a pseudonym used by David Griffin in a number of places.

A high performance Pygame OpenGL library, capable of handling 1000s of sprites

Requirements: Python, Pygame, PyOpenGL, Open GL 2.0 graphics hardware

Licensed under GPLv2 or GPLv3. You should have received a copy of both of these
licenses in the files GPLv2.txt and GPLv3.txt. If not these licenses can be
found at http://www.gnu.org/licenses/gpl-2.0.html (GPLv2) and
http://www.gnu.org/licenses/gpl-3.0.html (GPLv3). By using this file you agree
to be bound by at least one of these licenses.

Other licenses, including commercial/proprietary licenses, can be arranged by 
contacting me via e-mail at habilain@gmail.com. If you do not have in 
writing an alternative license agreement from me, you must use one of the 
licenses specified above when releasing work which makes use of this library.

pygl3display.sprites.attribsprite

Provides a version of the Sprite class that can be controlled by setting
attributes
"""

from pygl3display.sprites.sprite import Sprite

class AttribSprite(Sprite):
    """An AttribSprite is a sprite with the feature that setting its attributes
    will call the relevant functions to set the sprite data properly.
    For example:
     "attribSprite.position = (0, 0)" == "gl3sprite.setPosition((0,0))" 
    
    However, the caveat is that each operation is slower (all attribute writes
    on the sprite incur an additional function call overhead; this can add up
    quite quickly given it also applies internally!)
    
    In summary, this type of sprite is easier to program but carries a
    performance penalty. 
    
    One important note: As x and y scale can be set independently, there is
    no real "good" value to return for overall scale. So if reading the
    scale attribute, you get the x scale. If you're not setting the x or y 
    scaling independently, this is always correct. If you are, then don't
    read from the scale attribute.
    
    Magic attributes are as follows:
    image    = a GL3 image to use as the sprites image
    layer    = sprite layer
    flipx    = if the sprite should be flipped in x axis
    flipy    = if the sprite should be flipped in y axis
    scale    = scale the sprite by this amount
    scalex   = scale the sprite in the x direction by this amount
    scaley   = scale the sprite in the y direction by this amount
    visible  = if the sprite should be visible or not
    rotation = rotate this sprite about its origin by this amount
    """
    
    __init__ = Sprite.__init__
    magicVars = {'position':Sprite.setPosition, 'layer':Sprite.setLayer,
                 'flipx':Sprite.flipX, 'flipy':Sprite.flipY, 
                 'scale':Sprite.setScale, 'scalex':Sprite.setScaleX,
                 'scaley':Sprite.setScaleY, 'camera':Sprite.setCamera,
                 'visibile':Sprite.setVisibility, 
                 'rotation':Sprite.setRotation, 
                 'image':Sprite.setImage, 'offsets':Sprite.setOffsets
                 }
    magicStore = {'position':'_position', 'layer':'_layer', 'flipx':'_flipx',
                  'flipy':'_flipy', 'scale':'_scalex', 'scalex':'_scalex',
                  'scaley':'_scaley', 'visible':'_visible', 
                  'rotation':'_rotation', 'camera':'_camera',
                  '_image':'_image', 'offsets':'_unmodOffsets'}
    
    def __setattr__(self, attr, val):
        """Used to implement 'magic' variables"""
        if attr in type(self).magicVars:
            type(self).magicVars[attr](self, val)
        else:
            object.__setattr__(self, attr, val)
    
    def __getattr__(self, attr):
        """Used to get the value of 'magic' variables"""
        if attr in type(self).magicStore:
            return getattr(self, type(self).magicStore[attr])
        else: 
            raise AttributeError(str(type(self)) + ' has no attribute ' + attr)
