"""
PyGL3Display
(C) David Griffin (2010-present)
"Habilain" is a pseudonym used by David Griffin in a number of places.

A high performance Pygame OpenGL library, capable of handling 1000s of sprites

Requirements: Python, Pygame, PyOpenGL, Open GL 2.0 graphics hardware

Licensed under GPLv2 or GPLv3. You should have received a copy of both of these
licenses in the files GPLv2.txt and GPLv3.txt. If not these licenses can be
found at http://www.gnu.org/licenses/gpl-2.0.html (GPLv2) and
http://www.gnu.org/licenses/gpl-3.0.html (GPLv3). By using this file you agree
to be bound by at least one of these licenses.

Other licenses, including commercial/proprietary licenses, can be arranged by 
contacting me via e-mail at habilain@gmail.com. If you do not have in 
writing an alternative license agreement from me, you must use one of the 
licenses specified above when releasing work which makes use of this library.

pygl3display.sprites.groups

This file provides the old, software driven, sprite group class. 
"""

from pygl3display.sprites.sprite import Sprite
__all__ = ['SpriteGroup']

class SpriteGroupMetaclass(type): 
    """Metaclass for SpriteGroups - but could be one day extracted for any
    grouping, because it deals with arbitary types"""
    
    def wrapFunc(cls, containedtype, funcname):
        """Assuming that the class is an iterable containing only
        containedtypes, generate a moderately high performance wrapper that
        applies the named function to everything in the iterable."""
        if hasattr(containedtype, funcname):
            func = getattr(containedtype, funcname)
            if callable(func):
                def wrapper(self, *args, **kwargs):
                    for sprite in self:
                        func(sprite, *args, **kwargs)
                setattr(cls, funcname, wrapper)

class SpriteGroup(object):
    """An object to group together Sprites, like a Pygame group.
    SpriteGroup present a similar API to a Sprite, with an operation
    applied to the group being applied to all contained sprites.
    
    Due to metaclass magic, all operators on a Sprite will work on a
    SpriteGroup. However, for the moment, some operations will produces odd
    results, most notably setOrigin.
    """
    __metaclass__ = SpriteGroupMetaclass
    
    def __init__(self, sprites=None):
        """Create a SpriteGroup
           Arguments:
                sprites: The sprite or sprites to be in the group initially
        """
        self._vertexArrays = None 
        self._spritesByAtlas = {}
        self._slots = set()
        self._dirtyAtlases = {}
        self._deletedRenderSlots = set()
        self._deletedSprites = set()
        self._sprites = set()
        
        if sprites is not None:
            for container in sprites: 
                self.add(container)
        
    def __iter__(self):
        """Make the SpriteGroup an iterable"""
        self.clean()
        return self._sprites.__iter__()
        
    def __len__(self):
        """Provide a way of determining the number of elements in the group"""
        return self._sprites.__len__()
        
    def __getattr__(self, key):
        """Try to wrap unknown function calls - any function defined on
        a Sprite will also work on a SpriteGroup, but some operations may
        have odd results (e.g. setOrigin)"""
        type(self).wrapFunc(Sprite, key)
        if hasattr(self, key):
            return getattr(self, key)
        else:
            return super(SpriteGroup, self).__getattribute__(key)
    
    def clean(self):
        """Remove any sprites which have been removed from the group"""
        if self._dirtyAtlases:
            for atlas in self._dirtyAtlases:
                self._spritesByAtlas[atlas][0].difference_update(
                                                  self._dirtyAtlases[atlas][0])
                self._spritesByAtlas[atlas][1].difference_update(
                                                  self._dirtyAtlases[atlas][1]) 
            self._slots.difference_update(self._deletedRenderSlots)
            self._sprites.difference_update(self._deletedSprites)
            self._dirtyAtlases.clear()
            self._deletedRenderSlots.clear()
            self._deletedSprites.clear()
            
    def addSprite(self, sprite):
        """Add a single sprite to the Group"""
        if self._vertexArrays is None:
            self._vertexArrays = sprite._vertexArrays
        elif sprite._vertexArrays != self._vertexArrays:
            raise Exception('Tried to add something to a group which' + \
                            ' does not share the groups vertexArrays')
        if sprite not in self._sprites:
            spriteAtlas = sprite._renderer
            if spriteAtlas not in self._spritesByAtlas:
                self._spritesByAtlas[spriteAtlas] = [set(), set()]
            self._spritesByAtlas[spriteAtlas][0].update(
                                                    sprite._rdata)
            self._spritesByAtlas[spriteAtlas][1].update(sprite._rslots)
            self._sprites.add(sprite)
            self._slots.update(sprite._rslots)
            spritePosX, spritePosY = sprite._position
            sprite._baseOffsets = [(offsetX+spritePosX, offsetY+spritePosY) 
                            for offsetX, offsetY in sprite._baseOffsets]
            sprite._position = (0, 0)
            sprite._vertexArrays.setPosition(sprite._slots, 
                                             sprite._position)
            sprite._vertexArrays.setOffsets(sprite._slots, 
                                            sprite._baseOffsets)
        
    def add(self, thing):
        """Add a sprite or iterable container of sprites to the group
        Arguments:
         container: the sprite or container to add"""
        if isinstance(thing, Sprite):
            self.addSprite(thing)
        else:
            try:
                for sprite in thing:
                    self.addSprite(thing)
            except TypeError:
                raise Exception('Unsure how to add %s to SpriteGroup' 
                                 % str(thing))

