"""
PyGL3Display
(C) David Griffin (2010-present)
"Habilain" is a pseudonym used by David Griffin in a number of places.

A high performance Pygame OpenGL library, capable of handling 1000s of sprites

Requirements: Python, Pygame, PyOpenGL, Open GL 2.0 graphics hardware

Licensed under GPLv2 or GPLv3. You should have received a copy of both of these
licenses in the files GPLv2.txt and GPLv3.txt. If not these licenses can be
found at http://www.gnu.org/licenses/gpl-2.0.html (GPLv2) and
http://www.gnu.org/licenses/gpl-3.0.html (GPLv3). By using this file you agree
to be bound by at least one of these licenses.

Other licenses, including commercial/proprietary licenses, can be arranged by 
contacting me via e-mail at habilain@gmail.com. If you do not have in 
writing an alternative license agreement from me, you must use one of the 
licenses specified above when releasing work which makes use of this library.

pygl3display.vertexarrays.vertexarraysmeta

Provides the metaclass for vertexarrays, which provides the features to 
metaprogram the specific setters on the vertexarrays objects.
"""

class VertexArraysMetaclass(type):
    """A Metaclass providing factory functionality for setters on shaders.
    Arguments to the functions:
      name: the name the function should be given
      docdtype: the type of data (e.g. position coords) for documentation
      arrayName: the name of the attribute of the array 
            (i.e. getattr(self, array) = array on an instance of the class)
      itemSize: the number of elements in each item
      itemLen (not for makeValSetter): the number of elements in an item
      offset: An offset to apply (i.e. if packing scaleX, scaleY into a 2 float
              array, then scaleX has offset 0, scaleY has offset 1)
      """
      
    def finaliseFunc(cls, name, doc, func):
        """Finalises a metaprogrammed function by assigning its name, docstring
        and assigning it to the class"""
        func.__name__ = name
        func.__doc__ = doc
        setattr(cls, name, func)
      
    def makeValSetterDyn(cls, name, docdtype, arrayName, itemSize, itemLen, 
                         offset):
        """Makes a single value setter for a dynamic VBO"""
        doc = "Sets each slot in slots to " + docdtype + " specified"
        def setter(self, slots, val):
            array = self.arraysR[arrayName]
            dirtyElems = self.dirtyRange[arrayName]
            dirtyElems[0].add(slots[0] * itemSize + offset)
            dirtyElems[1].add(slots[-1] * itemSize + offset)
            self.dirty.add(arrayName)
            for pos in slots:
                array[pos*itemSize+offset] = val
        
        cls.finaliseFunc(name, doc, setter)
    
    def makeValSetter(cls, name, docdtype, arrayName, itemSize, itemLen, 
                      offset):
        """Makes a single value setter for a static VBO"""
        doc = "Sets each slot in slots to " + docdtype + " specified"
        def setter(self, slots, val):
            array = self.arraysR[arrayName]
            self.dirty.add(arrayName)
            for pos in slots:
                array[pos*itemSize+offset] = val
        
        cls.finaliseFunc(name, doc, setter)

    def makeVariableValSetterDyn(cls, name, docdtype, arrayName, itemSize, 
                              itemLen, offset):
        """Makes a single value setter that takes an offset for a dynamic 
        VBO"""
        doc = "Sets one of the " + docdtype + " of slots specified, as" + \
              "given by offset"
        def setter(self, slots, offset, val):
            array = self.arraysR[arrayName]
            dirtyElems = self.dirtyRange[arrayName]
            dirtyElems[0].add(slots[0] * itemSize + offset)
            dirtyElems[1].add(slots[-1] * itemSize + offset)
            self.dirty.add(arrayName)
            for pos in slots:
                array[pos*itemSize+offset] = val
            
        cls.finaliseFunc(name, doc, setter)
        
    def makeVariableValSetter(cls, name, docdtype, arrayName, itemSize, 
                              itemLen, offset):
        """Makes a single value setter that takes an offset for a static VBO"""
        doc = "Sets one of the " + docdtype + " of slots specified, as" + \
              "given by offset"
        def setter(self, slots, offset, val):
            array = self.arraysR[arrayName]
            self.dirty.add(arrayName)
            for pos in slots:
                array[pos*itemSize+offset] = val
            
        cls.finaliseFunc(name, doc, setter)
        
    def makeSwapper(cls, funcName, docstring, swaps):
        """Make an array swapper"""
        def swapper(self):
            for oldName, newName in swaps:
                temp = self.arraysR[oldName]
                self.arraysR[oldName] = self.arraysR[newName]
                self.arraysR[newName] = temp
        
        cls.finaliseFunc(funcName, docstring, swapper)

class VertexArraysMetaclassCtypes(VertexArraysMetaclass):

    def makeSetterDyn(cls, name, docdtype, arrayName, itemSize, itemLen, 
                      offset):
        """Make a tuple value setter for a dynamic VBO"""
        doc = "Sets each slot in slots to " + docdtype + " specified"
        itemLenPlusOffset = itemLen + offset
        def setter(self, slots, val):
            array = self.arraysR[arrayName]
            dirtyElems = self.dirtyRange[arrayName]
            dirtyElems[0].add(slots[0] * itemSize + offset)
            dirtyElems[1].add(slots[-1] * itemSize + itemLenPlusOffset)
            self.dirty.add(arrayName)
            for pos in slots:
                array[pos*itemSize+offset:pos*itemSize+itemLenPlusOffset] = val
        
        cls.finaliseFunc(name, doc, setter)
        
    def makeSetter(cls, name, docdtype, arrayName, itemSize, itemLen, offset):
        """Make a tuple value setter for a static VBO"""
        doc = "Sets each slot in slots to " + docdtype + " specified"
        itemLenPlusOffset = itemLen + offset
        def setter(self, slots, val):
            array = self.arraysR[arrayName]
            self.dirty.add(arrayName)
            for pos in slots:
                array[pos*itemSize+offset:pos*itemSize+itemLenPlusOffset] = val
        
        cls.finaliseFunc(name, doc, setter)
    
    def makeBatchSetterDyn(cls, name, docdtype, arrayName, itemSize, itemLen, 
                        offset): 
        """Makes a batch tuple setter for a dynamic VBO"""
        doc = ("Set each slot to the " + docdtype + " value corresponding to" +
               " its position in vals")
        lenplusOffset = itemLen+offset
        def batchSetter(self, slots, vals):
            array = self.arraysR[arrayName]
            dirtyElems = self.dirtyRange[arrayName]
            dirtyElems[0].add(slots[0] * itemSize + offset)
            dirtyElems[1].add(slots[-1] * itemSize + lenplusOffset)
            self.dirty.add(arrayName)
            for pos, val in zip(slots, vals):
                pos = pos*itemSize
                array[pos+offset:pos+lenplusOffset] = val
            
        cls.finaliseFunc(name, doc, batchSetter)
        
    def makeBatchSetter(cls, name, docdtype, arrayName, itemSize, itemLen, 
                        offset): 
        """Makes a batch tuple setter for a static VBO"""
        doc = ("Set each slot to the " + docdtype + " value corresponding to" +
               " its position in vals")
        lenplusOffset = itemLen+offset
        def batchSetter(self, slots, vals):
            array = self.arraysR[arrayName]
            self.dirty.add(arrayName)
            for pos, val in zip(slots, vals):
                pos = pos*itemSize
                array[pos+offset:pos+lenplusOffset] = val
            
        cls.finaliseFunc(name, doc, batchSetter)
        
class VertexArraysMetaclassNoSlices(VertexArraysMetaclass):
    """Provides a metaclass that does not rely on slicing. This is necessary
    until cffi implements either slicing on CData, or some other thing that
    could be used to improve performance"""
    def makeSetterDyn(cls, name, docdtype, arrayName, itemSize, itemLen, 
                      offset):
        """Make a tuple value setter for a dynamic VBO"""
        doc = "Sets each slot in slots to " + docdtype + " specified"
        itemLenPlusOffset = itemLen + offset
        def setter(self, slots, val):
            array = self.arraysR[arrayName]
            dirtyElems = self.dirtyRange[arrayName]
            dirtyElems[0].add(slots[0] * itemSize + offset)
            dirtyElems[1].add(slots[-1] * itemSize + itemLenPlusOffset)
            self.dirty.add(arrayName)
            for pos in slots:
                itemBase = pos * itemSize + offset 
                for x in xrange(itemLen):
                    array[itemBase + x] = val[x]
        
        cls.finaliseFunc(name, doc, setter)
        
    def makeSetter(cls, name, docdtype, arrayName, itemSize, itemLen, offset):
        """Make a tuple value setter for a static VBO"""
        doc = "Sets each slot in slots to " + docdtype + " specified"
        def setter(self, slots, val):
            array = self.arraysR[arrayName]
            self.dirty.add(arrayName)
            for pos in slots:
                itemBase = pos * itemSize + offset
                for x in xrange(itemLen):
                    array[itemBase + x] = val[x]
        
        cls.finaliseFunc(name, doc, setter)
    
    def makeBatchSetterDyn(cls, name, docdtype, arrayName, itemSize, itemLen, 
                        offset): 
        """Makes a batch tuple setter for a dynamic VBO"""
        doc = ("Set each slot to the " + docdtype + " value corresponding to" +
               " its position in vals")
        lenplusOffset = itemLen+offset
        def batchSetter(self, slots, vals):
            array = self.arraysR[arrayName]
            dirtyElems = self.dirtyRange[arrayName]
            dirtyElems[0].add(slots[0] * itemSize + offset)
            dirtyElems[1].add(slots[-1] * itemSize + lenplusOffset)
            self.dirty.add(arrayName)
            for pos, val in zip(slots, vals):
                itemBase = pos * itemSize + offset
                for x in xrange(itemLen):
                    array[itemBase+x] = val[x]
            
        cls.finaliseFunc(name, doc, batchSetter)
        
    def makeBatchSetter(cls, name, docdtype, arrayName, itemSize, itemLen, 
                        offset): 
        """Makes a batch tuple setter for a static VBO"""
        doc = ("Set each slot to the " + docdtype + " value corresponding to" +
               " its position in vals")
        def batchSetter(self, slots, vals):
            array = self.arraysR[arrayName]
            self.dirty.add(arrayName)
            for pos, val in zip(slots, vals):
                itemBase = pos * itemSize + offset
                for x in xrange(itemLen):
                    array[itemBase + x] = val[x]
            
        cls.finaliseFunc(name, doc, batchSetter)
        
