"""
PyGL3Display
(C) David Griffin (2010-present)
"Habilain" is a pseudonym used by David Griffin in a number of places.

A high performance Pygame OpenGL library, capable of handling 1000s of sprites

Requirements: Python, Pygame, PyOpenGL, Open GL 2.0 graphics hardware

Licensed under GPLv2 or GPLv3. You should have received a copy of both of these
licenses in the files GPLv2.txt and GPLv3.txt. If not these licenses can be
found at http://www.gnu.org/licenses/gpl-2.0.html (GPLv2) and
http://www.gnu.org/licenses/gpl-3.0.html (GPLv3). By using this file you agree
to be bound by at least one of these licenses.

Other licenses, including commercial/proprietary licenses, can be arranged by 
contacting me via e-mail at habilain@gmail.com. If you do not have in 
writing an alternative license agreement from me, you must use one of the 
licenses specified above when releasing work which makes use of this library.

pygl3display.vertexarrays.vertexarraysobj

Provides the vertex arrays objects; this is only a basic template, and a
specific version should be constructed using a VertexArraySpec
"""

import weakref
from pygl3display.glcommoninst import GLCommon
from pygl3display.slices import Slices
from pygl3display.gl import (glGenBuffers, glBufferData, glBufferSubData, 
    GL_UNSIGNED_INT, GL_ARRAY_BUFFER,  GL_STATIC_DRAW, GL_DYNAMIC_DRAW, 
    cdata, sizeOf, offsetPointer, byref, glGetError)
from pygl3display.vertexarrays.vertexarraysmeta import (
    VertexArraysMetaclassCtypes, VertexArraysMetaclassNoSlices)

class GenStore(object):
    def __init__(self, genFunc):
        self.genFunc = genFunc
        self.free = set()
        self.bound = {}
        
    def getOne(self, user):
        return self.get(user, 1)[0]
    
    def get(self, user, n):
        ret = []
        while self.free and n > 0:
            ret.append(self.free.pop())
            n -= 1
        if n > 0:
            ret.extend(self.genFunc(n))
        userref = weakref.ref(user, self.__release)
        self.bound[userref] = ret
        return ret
    
    def __release(self, userref):
        self.free.update(self.bound[userref])
        del self.bound[userref]
        
Buffers = GenStore(glGenBuffers) 

class VertexSlots(object):
    def __init__(self, slots, vertexArrays):
        self.slots = slots
        self.vertexArrays = vertexArrays
        
    def __iter__(self):
        return iter(self.slots)
    
    def __getitem__(self, indx):
        return self.slots.__getitem__(indx)
    
    
class VertexArrays(object):
    """Handles Vertex arrays, so that they can be shared amongst
    shaders as necessary. This class should not be initialised directly;
    instead, use VertexArraySpec to create a subclass which has all the
    appropriate setter methods already created."""
    __metaclass__ = VertexArraysMetaclassNoSlices
    
    def __init__(self, noSlots, arrays):
        """Initialise the vertex array container
        Arguments:
         noSlots: The total number of slots (i.e. length of the arrays)
         arrays : A Dictionary mapping array names (as in the shader program)
                  to a list containing the array itself as first element,
                  and as second element a list of the arguments to 
                  glVertexAttribPointer, without the first (location) or 
                  last (the actual array) argument
                  e.g. [someCTypesArray, [4, GL_FLOAT, GL_FALSE, 0]]"""
        
        self.positions = Slices(0, noSlots)
        self.arrays = arrays
        self.arraysR = dict([(key, self.arrays[key][0]) 
                              for key in self.arrays])
        self.shaderBindArgs = weakref.WeakKeyDictionary()
        self.dirty = set()
        self.dirtyRange = dict([(key, (set(), set())) for key in self.arrays])
        self.noSlots = noSlots
        self.indicesType = GL_UNSIGNED_INT
        self.indicesCType = cdata['I']
        self.sizeOfIndices = 4
        #buffers = glGenBuffers(len(arrays))
        buffers = Buffers.get(self, len(arrays))
        self.arrayBuffers = dict(zip(arrays.keys(), buffers))
        for key in self.arrayBuffers:
            bufferID = self.arrayBuffers[key]
            data = self.arraysR[key]
            GLCommon.bindBuffer(GL_ARRAY_BUFFER, bufferID)
            bufferType = (GL_DYNAMIC_DRAW if self.arrays[key][2] 
                          else GL_STATIC_DRAW)
            glBufferData(GL_ARRAY_BUFFER, len(data), data, bufferType)
        self.vertexArrayRefs = {}
                          
    def bind(self, force=False):
        """Bind the arrays as necessary, using the currently bound shader
        to determine locations"""
        shader = GLCommon.boundShader
        if shader not in self.shaderBindArgs:
            bindArgs = {}
            for key in self.arrays:
                if key in shader.arrayLocs:
                    bindArrayArgs = ((shader.arrayLocs[key],) + 
                        self.arrays[key][1] + (offsetPointer(0),))
                    bindArgs[key] = bindArrayArgs
            self.shaderBindArgs[shader] = bindArgs
        bindArgs = self.shaderBindArgs[shader]
        if GLCommon.boundVertexArrays is not self or force:
            GLCommon.boundVertexArrays = self
            bind = bindArgs
        else:
            bind = self.dirty

        for array in bind:
            bufferID = self.arrayBuffers[array]
            data = self.arraysR[array]
            dynamic = self.arrays[array][2]
            GLCommon.bindBuffer(GL_ARRAY_BUFFER, bufferID)
            if array in self.dirty:
                if dynamic:
                    dirtyElems = self.dirtyRange[array]
                    mn, mx = min(dirtyElems[0]), max(dirtyElems[1])
                    mult = int(sizeOf(data) / len(data))
                    sz = (mx - mn + 1) * mult
                    ref = byref(data, mn)
                    mn *= mult
                    glBufferSubData(GL_ARRAY_BUFFER, mn, sz, ref)
                    dirtyElems[0].clear()
                    dirtyElems[1].clear()
                else:
                    glBufferData(GL_ARRAY_BUFFER, len(data), data, 
                                 GL_STATIC_DRAW)
            GLCommon.bindArray(bindArgs[array])
        self.dirty.clear()
        
    def canAllocDataSlots(self, no):
        """Returns True if the VertexArrays have no or more slots remaining"""
        return self.positions.canAllocate(no) 
        
    def getDataSlots(self, no):
        """Get slots for storing vertex information
        Arguments:
         no: The number of slots to get"""
        slots = self.positions.allocate(no)
        ret = VertexSlots(slots, self)
        slotsRef = weakref.ref(ret, self.releaseDataSlots)
        self.vertexArrayRefs[slotsRef] = slots
        return ret
        
    def releaseDataSlots(self, slotsRef):
        """Release slots to be reused
        Arguments:
         slots: The slots to be released"""
        if slotsRef in self.vertexArrayRefs:
            self.positions.release(self.vertexArrayRefs[slotsRef])
            del self.vertexArrayRefs[slotsRef]

