"""
PyGL3Display
(C) David Griffin (2010-present)
"Habilain" is a pseudonym used by David Griffin in a number of places.

A high performance Pygame OpenGL library, capable of handling 1000s of sprites

Requirements: Python, Pygame, PyOpenGL, Open GL 2.0 graphics hardware

Licensed under GPLv2 or GPLv3. You should have received a copy of both of these
licenses in the files GPLv2.txt and GPLv3.txt. If not these licenses can be
found at http://www.gnu.org/licenses/gpl-2.0.html (GPLv2) and
http://www.gnu.org/licenses/gpl-3.0.html (GPLv3). By using this file you agree
to be bound by at least one of these licenses.

Other licenses, including commercial/proprietary licenses, can be arranged by 
contacting me via e-mail at habilain@gmail.com. If you do not have in 
writing an alternative license agreement from me, you must use one of the 
licenses specified above when releasing work which makes use of this library.

pygl3display.vertexarrays.vertexarrayspec

Provides the VertexArraySpec objects, which is how a vertexarray object 
should be constructed.
"""

from pygl3display.gl import (GL_FLOAT, GL_FALSE, cdata)
from pygl3display.vertexarrays.vertexarraysobj import VertexArrays

class VertexArraySpec(object):
    """Defines a specification for a class of vertex arrays"""
    def __init__(self, name, slotMultiplier=1.0):
        """Initialise the VertexArraySpec
        Args:
          name: The name of the class of vertex arrays to create
                (as would be used in documentation)
          slotMultiplier: A multiplier to give an idea of how many render
                          slots are needed per data point. For example, on a
                          sprite 4 data points are needed per sprite, but to
                          draw two triangles 6 render points are needed, so
                          the multiplier is 1.5. This is only a guide, can be
                          ignored if OpenGL wouldn't support a long enough 
                          array, and shouldn't result in too big a performance
                          hit even if it is completely ignored. Default: 1.0
        """
        self.name = name
        self.spec = {}
        self.attribs = []
        self.swappers = []
        self.vertexArraysClass = None
        self.slotMultiplier = slotMultiplier
        
    def copy(self, newName):
        """Return a copy of the vertex arrays spec, but this copy
        will always be unfrozen"""
        copy = type(self)(newName, self.slotMultiplier)
        copy.spec = self.spec.copy()
        copy.attribs = self.attribs[:]
        copy.swappers = self.swappers[:]
        return copy
        
    def addArray(self, name, cType=cdata['f'], glType=GL_FLOAT, 
                       noAttribs=4, glNorm=GL_FALSE, glStride=0,
                       dynamic=True):
        """Add an array to the specification. Defaults to float vec4s"""
        if self.vertexArraysClass is None:
            self.spec[name] = (cType, (noAttribs, glType, glNorm, glStride),
                                dynamic)
        else:
            raise Exception('VertexArraySpec has been frozen')
    
    def addAttrib(self, arrayName, attribName, attribSize, attribOffset,
                  attribBatched=False):
        """Add an attribute which will gain a setter function
        If attribOffset is None, the offset is an argument to the setter 
        function
        The functions added to the class are as follows:
        (with attribName == 'Attrib'; replace Attrib with the actual
         attribName for an actual usage')
        setAttribs(slots, values): 
          take two lists, slots and values, and set the value of slot[x] to
          values[x]
        setAttrib(slots, value, offset): 
          For each slot in slots, set the value of slot + offset to value
        setAttrib(slots, value): 
          For each slot in slots, set the value of the slot to value"""
        if self.vertexArraysClass is None:
            self.attribs.append((attribName, arrayName, attribSize, 
                                 attribOffset, attribBatched))
        else:
            raise Exception('VertexArraySpec has become frozen')
            
    def addSwapper(self, name, docstring, swaps):
        """Add an array swapper to the spec"""
        if self.vertexArraysClass is None:
            self.swappers.append((name, docstring, swaps))
        else:
            raise Exception('VertexArraySpec has become frozen')
            
    # Index format for setterMaker keys:
    # (attribBatched, singleAttrib, offsetIsNone, dynamic)
    setterMakers = {
        (True, False, False, False): 'makeBatchSetter',
        (False, True, True, False): 'makeVariableValSetter',
        (False, True, False, False): 'makeValSetter',
        (False, False, False, False): 'makeSetter',
        (True, False, False, True): 'makeBatchSetterDyn',
        (False, True, True, True): 'makeVariableValSetterDyn',
        (False, True, False, True): 'makeValSetterDyn',
        (False, False, False, True): 'makeSetterDyn',
    }            
    
    def makeVertexClass(self):
        """Create and metaprogram a class that matches the current spec"""
        class VertexArrayClass(VertexArrays):
            pass
        self.vertexArraysClass = VertexArrayClass
        VertexArrayClass.__name__ = self.name
        VertexArrayClass.__doc__ = "Stores and manages vertex arrays"""
        VertexArrayClass.slotMultiplier = self.slotMultiplier
        for attribName, arrayName, attribSize, attribOffset, attribBatched \
        in self.attribs:
            itemSize = self.spec[arrayName][1][0]
            setterName = 'set' + attribName
            if attribBatched:
                setterName += 's'
            offsetIsNone = attribOffset is None
            singleAttrib = attribSize == 1
            dynamic = self.spec[arrayName][2]
            
            signature = (attribBatched, singleAttrib, offsetIsNone, dynamic)
            if signature not in self.setterMakers:
                raise Exception('No valid setter maker')
            
            setterMakerName = self.setterMakers[signature]
            setterMaker = getattr(VertexArrayClass, setterMakerName)
            setterMaker(setterName, attribName, arrayName, itemSize, 
                        attribSize, attribOffset)
                                
        for swapperArgs in self.swappers:
            VertexArrayClass.makeSwapper(*swapperArgs)
            
    def makeArrays(self, length):
        """Freeze the spec and return a VertexArrays object of the class
        defined by the spec."""
        if self.vertexArraysClass is None:
            self.makeVertexClass()
        arrays = {}
        for name in self.spec:
            arraySpec = self.spec[name]
            cArray = arraySpec[0](arraySpec[1][0] * length)
            arrays[name] = [cArray, arraySpec[1], arraySpec[2]] 
        return self.vertexArraysClass(length, arrays)
